import argparse
import os
import typing
import urllib.request
import urllib.error
import json
import zipfile

GH_API = 'https://api.github.com'
GH_API_VERSION = '2022-11-28'
GH_API_ACCEPT = 'application/vnd.github.v3+json'
GH_API_VERSION_HEADER = 'X-GitHub-Api-Version'


class Params(typing.NamedTuple):
    owner: str
    repo: str
    version: str
    name: str
    token: str


class Response(typing.NamedTuple):
    status_code: int
    body: str

    def json(self) -> typing.Any:
        try:
            return json.loads(self.body)
        except json.JSONDecodeError:
            return None


def parse_args() -> Params:
    parser = argparse.ArgumentParser(
        description='Download artifact from GitHub Actions')
    parser.add_argument('--owner', type=str, required=True)
    parser.add_argument('--repo', type=str, required=True)
    parser.add_argument('--version', type=str, required=True,
                        help='Tag or short commit SHA')
    parser.add_argument('--name', type=str, required=True,
                        help='Artifact name')
    parser.add_argument('--token', type=str, required=True,
                        help='GitHub Personal Access Token with repo read permissions')

    args = parser.parse_args()
    return Params(
        owner=args.owner,
        repo=args.repo,
        version=args.version,
        name=args.name,
        token=args.token
    )


def execute_request(url: str, headers: dict) -> Response:
    print(f'Executing request: {url}')
    request = urllib.request.Request(
        url, headers=headers
    )

    try:
        with urllib.request.urlopen(request) as response:
            return Response(
                status_code=response.status,
                body=response.read().decode('utf-8')
            )
    except urllib.error.HTTPError as e:
        return Response(
            status_code=e.code,
            body=str(e.reason)
        )


def get_github_headers(token: str) -> dict:
    return {
        'Accept': GH_API_ACCEPT,
        GH_API_VERSION_HEADER: GH_API_VERSION,
        'Authorization': f'Bearer {token}',
    }


def get_artifact_url(params: Params) -> str:
    url = f'{GH_API}/repos/{params.owner}/{params.repo}/actions/artifacts?per_page=100'
    headers = get_github_headers(params.token)
    response = execute_request(url, headers)

    if response.status_code != 200:
        raise Exception(f'Failed to get artifacts: {response.body}')

    artifacts = response.json()
    for artifact in artifacts['artifacts']:
        sha: str = artifact['workflow_run']['head_sha']
        if sha.startswith(params.version) and artifact['name'] == params.name:
            return artifact['archive_download_url']

    raise Exception(
        f'Artifact "{params.name}" for version "{params.version}" not found'
    )


def get_release_url(params: Params) -> str:
    url = f'{GH_API}/repos/{params.owner}/{params.repo}/releases/tags/{params.version}'
    headers = get_github_headers(params.token)
    response = execute_request(url, headers)

    if response.status_code != 200:
        raise Exception(f'Failed to get release: {response.body}')

    release = response.json()
    for asset in release['assets']:
        if asset['name'] == params.name:
            return asset['url']

    raise Exception(
        f'Asset "{params.name}" for version "{params.version}" not found'
    )


def download_artifact(params: Params) -> None:
    headers = {
        'Authorization': f'token {params.token}',
    }

    # Check if specified version is a tag.
    # It is enought to check if it starts with 'v' character, since tags are
    # always prefixed with 'v', and commit SHAs cannot use 'v'.
    if params.version[0] == 'v':
        is_release = True
        url = get_release_url(params)
        headers['Accept'] = 'application/octet-stream'
    else:
        is_release = False
        url = get_artifact_url(params)

    request = urllib.request.Request(
        url, headers=headers
    )

    try:
        with urllib.request.urlopen(request) as response:
            if response.status != 200:
                raise Exception(f'Failed to download artifact: {response.reason}')

            # If file is not a release, it is a zip archive.
            name = params.name
            if not is_release:
                name += '.zip'

            with open(name, 'wb') as f:
                f.write(response.read())
    except urllib.error.HTTPError as e:
        raise Exception(f'Failed to download artifact: {e.reason}')

    # If file is not a release, it is a zip archive, so we need to extract it
    # and delete the zip archive.
    if not is_release:
        with zipfile.ZipFile(name) as zip_ref:
            zip_ref.extractall('.')
        os.remove(name)


def main() -> None:
    params = parse_args()
    download_artifact(params)


if __name__ == '__main__':
    main()
